// NESColExt.cpp : Implementation of CNESColExt

#include "stdafx.h"
#include "resource.h"
#include "NESColumns.h"
#include "NESColExt.h"

/////////////////////////////////////////////////////////////////////////////
// CNESColExt

STDMETHODIMP CNESColExt::GetColumnInfo (DWORD dwIndex, SHCOLUMNINFO* psci)
{
	// We have 3 columns: Mapper, PRG size, and CHR size
	if (dwIndex >= 3)
		return S_FALSE;

	switch (dwIndex)
	{
	case 0:	// Mapper
		psci->scid.fmtid = CLSID_NESColExt;	// Use our CLSID as the format ID
		psci->scid.pid   = 0;			// Use the col # as the ID
		psci->vt         = VT_LPSTR;		// We'll return the data as a string
		psci->fmt        = LVCFMT_RIGHT;	// Text will be right-aligned in the column
		psci->csFlags    = SHCOLSTATE_TYPE_INT; // Data should be sorted as numbers
		psci->cChars     = 3;			// Default col width in chars

		wcsncpy(psci->wszTitle, L"Mapper", MAX_COLUMN_NAME_LEN);
		wcsncpy(psci->wszDescription, L"iNES Mapper Number", MAX_COLUMN_DESC_LEN);
		psci->wszTitle[MAX_COLUMN_NAME_LEN-1] = 0;
		psci->wszDescription[MAX_COLUMN_DESC_LEN-1] = 0;
		break;

	case 1:	// PRG Size
		psci->scid.fmtid = CLSID_NESColExt;	// Use our CLSID as the format ID
		psci->scid.pid   = 1;			// Use the col # as the ID
		psci->vt         = VT_LPSTR;		// We'll return the data as a string
		psci->fmt        = LVCFMT_RIGHT;	// Text will be right-aligned in the column
		psci->csFlags    = SHCOLSTATE_TYPE_INT; // Data should be sorted as numbers
		psci->cChars     = 6;			// Default col width in chars

		wcsncpy(psci->wszTitle, L"PRG", MAX_COLUMN_NAME_LEN);
		wcsncpy(psci->wszDescription, L"PRG ROM Size", MAX_COLUMN_DESC_LEN);
		psci->wszTitle[MAX_COLUMN_NAME_LEN-1] = 0;
		psci->wszDescription[MAX_COLUMN_DESC_LEN-1] = 0;
		break;

	case 2:	// CHR Size
		psci->scid.fmtid = CLSID_NESColExt;	// Use our CLSID as the format ID
		psci->scid.pid   = 2;			// Use the col # as the ID
		psci->vt         = VT_LPSTR;		// We'll return the data as a string
		psci->fmt        = LVCFMT_RIGHT;	// Text will be right-aligned in the column
		psci->csFlags    = SHCOLSTATE_TYPE_INT; // Data should be sorted as numbers
		psci->cChars     = 6;			// Default col width in chars

		wcsncpy(psci->wszTitle, L"CHR", MAX_COLUMN_NAME_LEN);
		wcsncpy(psci->wszDescription, L"CHR ROM Size", MAX_COLUMN_DESC_LEN);
		psci->wszTitle[MAX_COLUMN_NAME_LEN-1] = 0;
		psci->wszDescription[MAX_COLUMN_DESC_LEN-1] = 0;
		break;

		DEFAULT_UNREACHABLE;
	}

	return S_OK;
}

STDMETHODIMP CNESColExt::GetItemData (LPCSHCOLUMNID pscid, LPCSHCOLUMNDATA pscd, VARIANT *pvarData )
{
	USES_CONVERSION;
	LPCTSTR	  szFilename = W2CT(pscd->wszFile);
	char	  szField[8];
	bool	  bUsingBuiltinCol = false;
	unsigned char rHeader[16];

	// Verify that the format id and column numbers are what we expect.
	if (pscid->fmtid == CLSID_NESColExt)
	{
		if (pscid->pid > 2)
			return S_FALSE;
	}
	else	return S_FALSE;

	// If we're being called with a directory (instead of a file), we can bail immediately.
	// Also bail if the file is offline (that is, backed up to tape or some other storage).	
	if (pscd->dwFileAttributes & (FILE_ATTRIBUTE_DIRECTORY | FILE_ATTRIBUTE_OFFLINE))
		return S_FALSE;

	// If the file's extension isn't ".NES", just skip it
	if (wcsicmp(pscd->pwszExt, L".nes"))
		return S_FALSE;

	if (!ReadNESHeader(szFilename, rHeader))
		return S_FALSE;

	// Check if we really have an iNES file by looking for the signature.
	if (strncmp((char *)rHeader, "NES\x1A", 4))
		return S_FALSE;

	// Do some other sanity checking
	if ((rHeader[7] & 0x0C) == 0x04)	// DiskDude!
		return S_FALSE;
	if ((rHeader[7] & 0x0C) == 0x0C)	// Not iNES 2.0
		return S_FALSE;

	int data;

	switch (pscid->pid)
	{
	case 0:	// Mapper number
		data = ((rHeader[6] & 0xF0) >> 4) | (rHeader[7] & 0xF0);
		if ((rHeader[7] & 0x0C) == 0x08)
			sprintf(szField, "%i.%i", data | ((rHeader[8] & 0x0F) << 8), (rHeader[8] & 0xF0) >> 4);
		else	sprintf(szField, "%i", data);
		break;

	case 1:	// PRG size
		data = rHeader[4];
                if ((rHeader[7] & 0x0C) == 0x08)
			data |= (rHeader[9] & 0x0F) << 8;
		sprintf(szField, "%iKB", data * 16);
		break;

	case 2:	// CHR size
		data = rHeader[5];
                if ((rHeader[7] & 0x0C) == 0x08)
			data |= (rHeader[9] & 0xF0) << 4;
		sprintf(szField, "%iKB", data * 8);
		break;

		DEFAULT_UNREACHABLE;
	}

	// Create a VARIANT with the details string, and return it back to the shell.
	CComVariant vData(szField);

	vData.Detach(pvarData);

	return S_OK;
}

//////////////////////////////////////////////////////////////////////////
//
// ReadNESHeader() : Reads an iNES header from a file.
//
// szFilename: Full path to the file.
// pTag: Pointer to a 16-byte buffer that receives the header
//
// Returns: true/false on success/failure.
//
// Note: This function does not verify that the header is actually correct,
//       it just reads it in and returns it
//
//////////////////////////////////////////////////////////////////////////

bool CNESColExt::ReadNESHeader (LPCTSTR szFilename, unsigned char *pHeader)
{
	HANDLE hfile;
	bool bReadLastAccessTime;
	FILETIME ftLastAccess;

	ATLASSERT(!IsBadStringPtr(szFilename, -1));
	ATLASSERT(!IsBadWritePtr(pHeader, 16));

	// Open the ROM file.
	hfile = CreateFile(szFilename, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);

	if (INVALID_HANDLE_VALUE == hfile)
		return false;

	// Save the last-accessed time so we can restore it later.  Note that when
	// testing on XP SP2 (and earlier on 2K), by the time this code is called,
	// the file's last-accessed time is already set to the current date/time,
	// so this code here doesn't have any effect.
	bReadLastAccessTime = (0 != GetFileTime ( hfile, NULL, &ftLastAccess, NULL ));

	__try
	{
		DWORD dwFileSizeLo, dwFileSizeHi, dwBytesRead;

		// Check the file length
		dwFileSizeLo = GetFileSize ( hfile, &dwFileSizeHi );

		// Did GetFileSize() succeed?
		if ( -1 == dwFileSizeLo && GetLastError() != ERROR_SUCCESS )
			return false;

		// Is the file large enough to be an iNES ROM image?
		if ( 0 == dwFileSizeHi && dwFileSizeLo < 16400 )
			return false;

		// Read in the header
		if (!ReadFile(hfile, pHeader, 16, &dwBytesRead, NULL))
			return false;

		if (dwBytesRead != 16)
			return false;
	}
	__finally
	{
		ATLASSERT(INVALID_HANDLE_VALUE != hfile);

		// Restore the file's last-accessed time, and close it.
		if (bReadLastAccessTime)
			SetFileTime(hfile, NULL, &ftLastAccess, NULL);

		CloseHandle(hfile);
	}

	return true;
}